@@ -25,7 +25,7 @@ discopy computes natural language meaning in pictures.
 
 
 from discopy import (
-    config, cat, monoidal, rigid, biclosed, tensor, quantum, grammar)
+    config, cat, monoidal, rigid, biclosed, tensor, matrix, quantum, grammar)
 from discopy.utils import from_tree
 from discopy.cat import Quiver
 from discopy.monoidal import Sum
@@ -33,7 +33,7 @@ from discopy.rigid import (
     Ob, Ty, PRO, Box, Diagram, Id, Cup, Cap, Swap, Functor)
 from discopy.tensor import Dim, Tensor, Functor as TensorFunctor
 
-from discopy.quantum import cqmap, zx
+from discopy.quantum import circuit, cqmap, zx, optics
 from discopy.quantum.zx import Z, X, H, SWAP
 from discopy.quantum.cqmap import C, Q, CQMap
 from discopy.quantum.circuit import (
@@ -46,4 +46,4 @@ from discopy.grammar import cfg, ccg, pregroup
 from discopy.grammar.pregroup import Word
 
 
-__version__ = '0.3.7.2'
+__version__ = '0.4.3'
