@@ -38,11 +38,13 @@ We can check the Eckmann-Hilton argument, up to interchanger.
 .. image:: ../_static/imgs/EckmannHilton.gif
     :align: center
 """
-
 from discopy import cat, messages, drawing, rewriting
 from discopy.cat import Ob
+from discopy.messages import WarnOnce
 from discopy.utils import factory_name, from_tree
 
+warn_permutation = WarnOnce()
+
 
 class Ty(Ob):
     """
@@ -541,10 +543,14 @@ class Diagram(cat.Arrow):
             >> ar_factory.swap(left[:1], right) @ ar_factory.id(left[1:])
 
     @staticmethod
-    def permutation(perm, dom=None, ar_factory=None):
+    def permutation(perm, dom=None, ar_factory=None, inverse=False):
         """
         Returns the diagram that encodes a permutation of wires.
 
+        .. warning::
+            This method used to return the inverse permutation up to and
+            including discopy v0.4.2.
+
         Parameters
         ----------
         perm : list of int
@@ -552,6 +558,8 @@ class Diagram(cat.Arrow):
         dom : monoidal.Ty, optional
             of the same length as :code:`perm`,
             default is :code:`PRO(len(perm))`.
+        inverse : bool
+            whether to return the inverse permutation.
 
         Returns
         -------
@@ -562,6 +570,12 @@ class Diagram(cat.Arrow):
             raise ValueError("Input should be a permutation of range(n).")
         if dom is None:
             dom = PRO(len(perm))
+        if not inverse:
+            warn_permutation.warn(
+                'Since discopy v0.4.3 the behaviour of '
+                'permutation has changed. Pass inverse=False '
+                'to get the default behaviour.')
+            perm = [perm.index(i) for i in range(len(perm))]
         if len(dom) != len(perm):
             raise ValueError(
                 "Domain and permutation should have the same length.")
@@ -574,7 +588,7 @@ class Diagram(cat.Arrow):
             perm = perm[:i] + [i] + perm[i:j] + perm[j + 1:]
         return diagram
 
-    def permute(self, *perm):
+    def permute(self, *perm, inverse=False):
         """
         Returns :code:`self >> self.permutation(perm, self.dom)`.
 
@@ -582,13 +596,24 @@ class Diagram(cat.Arrow):
         ----------
         perm : list of int
             such that :code:`i` goes to :code:`perm[i]`
+        inverse : bool
+            whether to return the inverse permutation.
 
         Examples
         --------
         >>> x, y, z = Ty('x'), Ty('y'), Ty('z')
         >>> assert Id(x @ y @ z).permute(2, 1, 0).cod == z @ y @ x
+        >>> assert Id(x @ y @ z).permute(2, 0).cod == z @ y @ x
         """
-        return self >> self.permutation(list(perm), self.dom)
+        if min(perm) < 0 or max(perm) >= len(self.cod):
+            raise IndexError(f'{self} index out of bounds.')
+        if len(set(perm)) != len(perm):
+            raise ValueError('{perm} is not a permutation.')
+        sorted_perm = sorted(perm)
+        perm = [
+            i if i not in perm else sorted_perm[perm.index(i)]
+            for i in range(len(self.cod))]
+        return self >> self.permutation(list(perm), self.cod, inverse)
 
     @staticmethod
     def subclass(ar_factory):
