@@ -26,7 +26,7 @@ def test_Ty_repr():
 
 
 def test_Ty_str():
-    str(Ty('x')) == 'x'
+    assert str(Ty('x')) == 'x'
 
 
 def test_Ty_getitem():
@@ -143,7 +143,10 @@ def test_Diagram_permutation():
     x = PRO(1)
     assert Diagram.swap(x, x ** 2)\
         == Diagram.swap(x, x) @ Id(x) >> Id(x) @ Diagram.swap(x, x)\
-        == Diagram.permutation([2, 0, 1])
+        == Diagram.permutation([2, 0, 1], inverse=True)
+    assert Diagram.swap(x, x ** 2)\
+        == Diagram.swap(x, x) @ Id(x) >> Id(x) @ Diagram.swap(x, x)\
+        == Diagram.permutation([2, 0, 1]).dagger()
     with raises(ValueError):
         Diagram.permutation([2, 0])
     with raises(ValueError):
@@ -349,3 +352,10 @@ def test_Sum():
     assert {Sum([f]): 42}[Sum([f])] == 42
     assert Id(x).then(*(3 * (f + f, ))) == sum(8 * [f >> f >> f])
     assert Id(Ty()).tensor(*(3 * (f + f, ))) == sum(8 * [f @ f @ f])
+
+
+def test_bad_permute():
+    with raises(IndexError):
+        Id(Ty('n')).permute(1)
+    with raises(ValueError):
+        Id(Ty('n')).permute(0, 0)
