@@ -1,14 +1,15 @@
 # -*- coding: utf-8 -*-
 
 from unittest.mock import Mock
-from functools import reduce, partial
-import itertools
+from functools import partial
 from pytest import raises
 import numpy as np
 from discopy.quantum.cqmap import *
 from discopy.quantum.circuit import *
 from discopy.quantum.gates import *
 from discopy.quantum import tk
+import torch
+import sympy
 
 
 def test_index2bitstring():
@@ -37,7 +38,10 @@ def test_Circuit_permutation():
     x = qubit
     assert Circuit.swap(x, x ** 2)\
         == Circuit.swap(x, x) @ Id(x) >> Id(x) @ Circuit.swap(x, x)\
-        == Circuit.permutation([2, 0, 1])
+        == Circuit.permutation([2, 0, 1], inverse=True)
+    assert Circuit.swap(x, x ** 2)\
+        == Circuit.swap(x, x) @ Id(x) >> Id(x) @ Circuit.swap(x, x)\
+        == Circuit.permutation([2, 0, 1]).dagger()
 
 
 def test_Circuit_eval():
@@ -55,23 +59,36 @@ def test_Circuit_cups_and_caps():
 
 def test_Circuit_spiders():
     assert Circuit.spiders(123, 456, qubit ** 0) == Id()
-    assert Circuit.spiders(0, 0, qubit) == Ket(0) >> H >> H >> Bra(0)
+    assert Circuit.spiders(0, 0, qubit) == (sqrt(2) >> Ket(0)
+                                            >> H >> H
+                                            >> Bra(0) >> sqrt(2))
     assert Circuit.spiders(1, 1, qubit) == Id(qubit)
-    assert Circuit.spiders(0, 1, qubit ** 2) == (Ket(0) >> H) @ (Ket(0) >> H)
+    assert Circuit.spiders(0, 1, qubit ** 2) == ((sqrt(2) >> Ket(0) >> H)
+                                                 @ (sqrt(2) >> Ket(0) >> H))
 
-    mul2 = Circuit(
-        dom=qubit @ qubit @ qubit @ qubit, cod=qubit @ qubit,
-        boxes=[SWAP, CX, Bra(0), CX, Bra(0)], offsets=[1, 0, 1, 1, 2])
+    mul2 = Circuit(dom=qubit @ qubit @ qubit @ qubit,
+                   cod=qubit @ qubit,
+                   boxes=[SWAP, CX, Bra(0), CX, Bra(0)],
+                   offsets=[1, 0, 1, 1, 2])
     assert Circuit.spiders(2, 1, qubit ** 2) == mul2
 
-    ghz2 = Circuit(
-        dom=Ty(), cod=qubit @ qubit @ qubit @ qubit @ qubit @ qubit,
-        boxes=[
-            Ket(0), H, Ket(0), CX, Ket(0), CX, Ket(0), H, Ket(0), CX, Ket(0),
-            CX, SWAP, SWAP, SWAP],
-        offsets=[0, 0, 1, 0, 1, 0, 3, 3, 4, 3, 4, 3, 2, 1, 3])
+    ghz2 = Circuit(dom=Ty(),
+                   cod=qubit @ qubit @ qubit @ qubit @ qubit @ qubit,
+                   boxes=[sqrt(2), Ket(0), H, Ket(0), CX, Ket(0), CX,
+                          sqrt(2), Ket(0), H, Ket(0), CX, Ket(0), CX,
+                          SWAP, SWAP, SWAP],
+                   offsets=[0, 0, 0, 1, 0, 1, 0, 3, 3, 3, 4, 3, 4,
+                            3, 2, 1, 3])
     assert Circuit.spiders(0, 3, qubit ** 2) == ghz2
 
+    assert np.abs(Circuit.spiders(0, 0, qubit).eval().array) == 2
+
+    combos = [(2, 3), (5, 4), (0, 1)]
+    for n_legs_in, n_legs_out in combos:
+        flat_tensor = np.abs(Circuit.spiders(n_legs_in, n_legs_out, qubit)
+                             .eval().array.flatten())
+        assert flat_tensor[0] == flat_tensor[-1] == 1
+
 
 def test_Circuit_to_tk():
     bell_state = Circuit.caps(qubit, qubit)
@@ -97,6 +114,141 @@ def test_Circuit_to_tk():
            ".post_select({1: 0}).post_process(Swap(bit, bit))"
 
 
+def test_Circuit_to_pennylane(capsys):
+    bell_state = Circuit.caps(qubit, qubit)
+    bell_effect = bell_state[::-1]
+    snake = (bell_state @ Id(1) >> Bra(0) @ bell_effect)[::-1]
+    p_snake = snake.to_pennylane()
+    p_snake.draw()
+
+    captured = capsys.readouterr()
+    assert captured.out == \
+        ("0: ───────╭●──H─┤0>\n"
+         "1: ──H─╭●─╰X────┤0>\n"
+         "2: ────╰X───────┤  State\n")
+
+    assert np.allclose(p_snake.eval().numpy(), snake.eval().array)
+
+    p_snake_prob = snake.to_pennylane(probabilities=True)
+    snake_prob = (snake >> Measure())
+
+    assert np.allclose(p_snake_prob.eval().numpy(), snake_prob.eval().array)
+
+    no_open_snake = (bell_state @ Ket(0) >> Bra(0) @ bell_effect)[::-1]
+    p_no_open_snake = no_open_snake.to_pennylane()
+    p_no_open_snake.draw()
+
+    captured = capsys.readouterr()
+    assert captured.out == \
+        ("0: ───────╭●──H─┤0>\n"
+         "1: ──H─╭●─╰X────┤0>\n"
+         "2: ────╰X───────┤0>\n")
+
+    assert np.allclose(p_no_open_snake.eval().numpy(),
+                       no_open_snake.eval().array)
+
+    # probabilities should not be normalized if all wires are post-selected
+    p_no_open_snake_prob = no_open_snake.to_pennylane(probabilities=True)
+
+    assert np.allclose(p_no_open_snake_prob.eval().numpy(),
+                       no_open_snake.eval().array)
+
+    x, y, z = sympy.symbols('x y z')
+    symbols = [x, y, z]
+    weights = [torch.tensor([1.]), torch.tensor([2.]), torch.tensor([3.])]
+
+    var_circ = Circuit(
+        dom=qubit ** 0, cod=qubit, boxes=[
+            Ket(0), Rx(0.552), Rz(x), Rx(0.917), Ket(0, 0, 0), H, H, H,
+            CRz(0.18), CRz(y), CX, H, sqrt(2), Bra(0, 0), Ket(0),
+            Rx(0.446), Rz(0.256), Rx(z), CX, H, sqrt(2), Bra(0, 0)],
+        offsets=[
+            0, 0, 0, 0, 0, 0, 1, 2, 0, 1, 2, 2,
+            3, 2, 0, 0, 0, 0, 0, 0, 1, 0])
+
+    p_var_circ = var_circ.to_pennylane()
+    p_var_circ.draw(symbols, weights)
+
+    captured = capsys.readouterr()
+    assert captured.out == \
+        ("0: ──RX(2.80)──RZ(1.61)──RX(18.85)─╭●──H─┤0>\n"
+         "1: ──H────────╭●───────────────────╰X────┤0>\n"
+         "2: ──H────────╰RZ(1.13)─╭●───────────────┤  State\n"
+         "3: ──H──────────────────╰RZ(12.57)─╭●──H─┤0>\n"
+         "4: ──RX(3.47)──RZ(6.28)──RX(5.76)──╰X────┤0>\n")
+
+    var_f = var_circ.lambdify(*symbols)
+    conc_circ = var_f(*[a.item() for a in weights])
+
+    assert np.allclose(p_var_circ.eval(symbols, weights).numpy(),
+                       conc_circ.eval().array)
+
+    p_var_circ_prob = var_circ.to_pennylane(probabilities=True)
+    conc_circ_prob = (conc_circ >> Measure())
+
+    assert (np.allclose(p_var_circ_prob.eval(symbols, weights).numpy(),
+                        conc_circ_prob.eval().array))
+
+
+def test_PennyLaneCircuit_mixed_error():
+    bell_state = Circuit.caps(qubit, qubit)
+    bell_effect = bell_state[::-1]
+    snake = (bell_state @ Id(1) >> Bra(0) @ bell_effect)[::-1]
+    snake = (snake >> Measure())
+    with raises(ValueError):
+        snake.to_pennylane()
+
+
+def test_PennylaneCircuit_draw(capsys):
+    bell_state = Circuit.caps(qubit, qubit)
+    bell_effect = bell_state[::-1]
+    snake = (bell_state @ Id(1) >> Bra(0) @ bell_effect)[::-1]
+    p_circ = snake.to_pennylane()
+    p_circ.draw()
+
+    captured = capsys.readouterr()
+    assert captured.out == \
+        ("0: ───────╭●──H─┤0>\n"
+         "1: ──H─╭●─╰X────┤0>\n"
+         "2: ────╰X───────┤  State\n")
+
+
+def test_pennylane_ops():
+    ops = [X, Y, Z, S, T, H, CX, CZ]
+
+    for op in ops:
+        disco = (Id().tensor(*([Ket(0)] * len(op.dom))) >> op).eval().array
+        plane = op.to_pennylane().eval().numpy()
+
+        assert np.allclose(disco, plane)
+
+
+def test_pennylane_parameterized_ops():
+    ops = [Rx, Ry, Rz, CRx, CRz]
+
+    for op in ops:
+        p_op = op(0.5)
+        disco = (Id().tensor(*([Ket(0)] * len(p_op.dom))) >> p_op).eval().array
+        plane = p_op.to_pennylane().eval().numpy()
+
+        assert np.allclose(disco, plane, atol=10e-5)
+
+
+def test_pennylane_update_post_selection():
+    bell_state = Circuit.caps(qubit, qubit)
+    bell_effect = bell_state[::-1]
+    snake = (bell_state @ Id(1) >> Bra(0) @ bell_effect)[::-1]
+    p_circ = snake.to_pennylane()
+
+    assert p_circ.post_selection == {0: 0, 1: 0}
+    assert p_circ._valid_states == [0, 1]
+
+    p_circ.post_selection = {0: 0, 2: 0}
+
+    assert p_circ.post_selection == {0: 0, 2: 0}
+    assert p_circ._valid_states == [0, 2]
+
+
 def test_Sum_from_tk():
     assert Circuit.from_tk(*(X + X).to_tk()) == (X + X).init_and_discard()
     assert Circuit.from_tk() == Sum([], qubit ** 0, qubit ** 0)
@@ -127,10 +279,14 @@ def test_Circuit_from_tk():
 
     m = Measure(1, destructive=False, override_bits=True)
     assert back_n_forth(m) == m.init_and_discard()
+    assert back_n_forth(CRx(0.5)) ==\
+        Ket(0) @ Ket(0) >> CRx(0.5) >> Discard() @ Discard()
     assert back_n_forth(CRz(0.5)) ==\
         Ket(0) @ Ket(0) >> CRz(0.5) >> Discard() @ Discard()
     assert Id(qubit @ bit).init_and_discard()\
         == back_n_forth(Swap(qubit, bit)) == back_n_forth(Swap(bit, qubit))
+    c = (T >> T.dagger()).init_and_discard()
+    assert c == back_n_forth(c)
 
 
 def test_ClassicalGate_to_tk():
@@ -141,6 +297,11 @@ def test_ClassicalGate_to_tk():
     assert Circuit.from_tk(circuit.to_tk())[-1] == post
 
 
+def test_tk_dagger():
+    assert S.dagger().to_tk() == tk.Circuit(1).Sdg(0)
+    assert T.dagger().to_tk() == tk.Circuit(1).Tdg(0)
+
+
 def test_Circuit_get_counts():
     assert Id(1).get_counts() == {(): 1.0}
 
@@ -252,7 +413,7 @@ def test_Measure():
 def test_QuantumGate():
     assert repr(X) == "X"
     assert repr(QuantumGate("s", 0, [1]))\
-        == "QuantumGate('s', n_qubits=0, array=[1])"
+        == "QuantumGate('s', n_qubits=0, array=[1.+0.j])"
 
 
 def test_ClassicalGate():
@@ -313,6 +474,16 @@ def test_IQPAnsatz():
         IQPansatz(10, np.array([]))
 
 
+def test_Sim14Ansatz():
+    with raises(ValueError):
+        Sim14ansatz(10, np.array([]))
+
+
+def test_Sim15Ansatz():
+    with raises(ValueError):
+        Sim15ansatz(10, np.array([]))
+
+
 def test_Sum():
     assert not Sum([], qubit, qubit).eval()
     assert Sum([Id(0)]).get_counts() == Id(0).get_counts()
@@ -380,21 +551,28 @@ def test_testing_utils():
 
 
 def test_rot_grad():
-    from sympy import I, pi
+    from sympy import pi
     from sympy.abc import phi
-    assert CRz(phi).grad(phi, mixed=False)\
-        == (CRz(phi) >> Z @ Z @ scalar(0.5 * I * pi))\
-        + (CRz(phi) >> Id(1) @ Z @ scalar(-0.5 * I * pi))
-    assert CRx(phi).grad(phi, mixed=False)\
-        == (CRx(phi) >> Z @ X @ scalar(0.5 * I * pi))\
-        + (CRx(phi) >> Id(1) @ X @ scalar(-0.5 * I * pi))
+    x = 0.7
+    crz_diff = CRz(phi).grad(phi, mixed=False).lambdify(phi)(x).eval()
+    crz_res = (
+        (CRz(phi) >> Z @ Z @ scalar(0.5j * pi))
+        + (CRz(phi) >> Id(1) @ Z @ scalar(-0.5j * pi))
+    ).lambdify(phi)(x).eval()
+    assert np.allclose(crz_diff, crz_res)
+
+    crx_diff = CRx(phi).grad(phi, mixed=False).lambdify(phi)(x).eval()
+    crx_res = (
+        (CRx(phi) >> Z @ X @ scalar(0.5j * pi))
+        + (CRx(phi) >> Id(1) @ X @ scalar(-0.5j * pi))
+    ).lambdify(phi)(x).eval()
+    assert np.allclose(crx_diff, crx_res)
 
 
 def test_rot_grad_NotImplemented():
     from sympy.abc import z
-    for gate in (CRx, CRz, CU1):
-        with raises(NotImplementedError):
-            gate(z).grad(z, mixed=True)
+    with raises(NotImplementedError):
+        CU1(z).grad(z, mixed=True)
 
 
 def test_ClassicalGate_grad_subs():
@@ -497,8 +675,8 @@ def test_real_amp_ansatz():
 def test_Controlled():
     with raises(TypeError):
         Controlled(None)
-    with raises(NotImplementedError):
-        Controlled(X, distance=-1)
+    with raises(ValueError):
+        Controlled(X, distance=0)
 
 
 def test_adjoint():
@@ -523,14 +701,76 @@ def test_adjoint():
     gates_conj = [
         Bra(0), Ket(0, 0), Rx(-0.1), Ry(0.2), Rz(-0.3),
         Swap(qubit, qubit) >> CU1(-0.4) >> Swap(qubit, qubit),
-        Swap(qubit, qubit) >> CRx(-0.5) >> Swap(qubit, qubit),
-        Swap(qubit, qubit) >> CRz(-0.7) >> Swap(qubit, qubit),
+        Controlled(Rx(-0.5), distance=-1),
+        Controlled(Rz(-0.7), distance=-1),
         Scalar(1 - 2j),
-        QuantumGate(
-            'CX', n_qubits=2, _conjugate=True,
-            array=[1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0]),
+        Controlled(X, distance=-1),
         Swap(qubit, bit), Copy(), Match()
     ]
 
     for g, g_conj in zip(gates, gates_conj):
         assert g.conjugate() == g_conj
+
+
+def test_causal_cx():
+    assert np.allclose((CX >> Discard(2)).eval(), Discard(2).eval())
+
+
+def test_eval_no_qutrits():
+    qutrit = Ty(Qudit(3))
+    with raises(Exception):
+        Box('qutrit box', qutrit, qutrit).to_tn(mixed=True)
+
+
+def test_grad_unknown_controlled():
+    from sympy.abc import phi
+    unknown = QuantumGate('gate', 1, data=phi)
+    with raises(NotImplementedError):
+        Controlled(unknown).grad(phi)
+
+
+def test_symbolic_controlled():
+    from sympy.abc import phi
+    crz = lambda x, d: Controlled(Rz(x), distance=d)
+    assert np.all(
+        crz(phi, -1).eval().array
+        == (SWAP >> crz(phi, 1) >> SWAP).eval().array)
+
+
+def test_controlled_subs():
+    from sympy.abc import phi, psi
+    assert CRz(phi).subs(phi, 0.1) == CRz(0.1)
+    assert CRx(psi).l.subs((phi, 0.1), (psi, 0.2)) == CRx(0.2).l
+
+
+def test_circuit_chaining():
+    circuit = Id(5).CX(0, 2).X(4).CRz(0.2, 4, 2).H(2)
+    expected_circuit = Circuit(
+        dom=qubit @ qubit @ qubit @ qubit @ qubit,
+        cod=qubit @ qubit @ qubit @ qubit @ qubit,
+        boxes=[
+            Controlled(X, distance=2), X, Controlled(Rz(0.2), distance=-2), H],
+        offsets=[0, 4, 2, 2])
+    assert circuit == expected_circuit
+
+    circuit = Id(3).CZ(0, 1).CX(2, 0).CCX(1, 0, 2).CCZ(2, 0, 1).X(0).Y(1).Z(2)
+    expected_circuit = Circuit(
+        dom=qubit @ qubit @ qubit, cod=qubit @ qubit @ qubit,
+        boxes=[
+            CZ, Controlled(X, distance=-2), Controlled(CX, distance=1),
+            Controlled(CZ, distance=-1), X, Y, Z],
+        offsets=[0, 0, 0, 0, 0, 1, 2])
+    assert circuit == expected_circuit
+
+    circuit = Id(1).Rx(0.1, 0).Ry(0.2, 0).Rz(0.3, 0)
+    expected_circuit = Rx(0.1) >> Ry(0.2) >> Rz(0.3)
+    assert circuit == expected_circuit
+
+    assert Id(1).S(0) == S
+
+    with raises(ValueError):
+        Id(3).CY(0, 0)
+    with raises(ValueError):
+        Id(1).CRx(0.7, 1, 0)
+    with raises(ValueError):
+        Id(2).X(999)
