@@ -1,17 +1,38 @@
-from pytest import raises
+from pytest import raises, fixture
 from discopy import *
 from discopy.quantum.zx import *
 import numpy as np
+import random
+
+
+@fixture
+def random_had_cnot_diagram():
+    def _random_had_cnot_diagram(qubits, depth, p_had=0.5):
+        random.seed(0)
+        c = Ket(*(qubits * [0]))
+        for _ in range(depth):
+            r = random.random()
+            if r > p_had:
+                c = c.H(random.randrange(qubits))
+            else:
+                tgt = random.randrange(qubits)
+                while True:
+                    ctrl = random.randrange(qubits)
+                    if ctrl != tgt:
+                        break
+                c = c.CX(tgt, ctrl)
+        return c
+    return _random_had_cnot_diagram
 
 
 def test_Diagram():
     bialgebra = Z(1, 2) @ Z(1, 2) >> Id(1) @ SWAP @ Id(1) >> X(2, 1) @ X(2, 1)
-    repr(bialgebra) == "zx.Diagram(dom=PRO(2), cod=PRO(2), "\
-                       "boxes=[Z(1, 2), Z(1, 2), SWAP, X(2, 1), X(2, 1)], "\
-                       "offsets=[0, 2, 1, 0, 1])"
-    str(bialgebra) == "Z(1, 2) @ Id(1) >> Id(2) @ Z(1, 2)"\
-                      ">> Id(1) @ SWAP @ Id(1)"\
-                      ">> X(2, 1) @ Id(2) >> Id(1) @ X(2, 1)"
+    assert repr(bialgebra) == "zx.Diagram(dom=PRO(2), cod=PRO(2), boxes=" \
+                              "[Z(1, 2), Z(1, 2), SWAP, X(2, 1), X(2, 1)], " \
+                              "offsets=[0, 2, 1, 0, 1])"
+    assert str(bialgebra) == "Z(1, 2) @ Id(1) >> Id(2) @ Z(1, 2) " \
+                             ">> Id(1) @ SWAP @ Id(1) " \
+                             ">> X(2, 1) @ Id(2) >> Id(1) @ X(2, 1)"
 
 
 def test_Swap():
@@ -76,22 +97,13 @@ def test_grad():
            + (Z(1, 1, phi / 2) >> scalar(pi) @ Id(1) >> Z(1, 1, phi + 1.5))
 
 
-def test_grad_to_pyzx():
-    from sympy.abc import theta
-    m1 = circuit2zx(CU1(theta).grad(theta, mixed=False))\
-        .subs(theta, 1 / 2).to_pyzx().to_matrix()
-    e3 = _std_basis_v(1, 1)
-    m2 = (e3 @ e3.T) * (-2j * np.pi)
-    assert np.isclose(np.linalg.norm(m1 - m2), 0)
-
-
 def test_to_pyzx_errors():
     with raises(TypeError):
         Diagram.to_pyzx(quantum.H)
 
 
 def test_to_pyzx():
-    Diagram.from_pyzx(Z(0, 2).to_pyzx()) == Z(0, 2) >> SWAP
+    assert Diagram.from_pyzx(Z(0, 2).to_pyzx()) == Z(0, 2) >> SWAP
 
 
 def test_to_pyzx_scalar():
@@ -105,15 +117,17 @@ def test_to_pyzx_scalar():
 def test_from_pyzx_errors():
     bialgebra = Z(1, 2) @ Z(1, 2) >> Id(1) @ SWAP @ Id(1) >> X(2, 1) @ X(2, 1)
     graph = bialgebra.to_pyzx()
-    graph.inputs = graph.outputs = []
+    graph.set_inputs(())
+    graph.set_outputs(())
     with raises(ValueError):  # missing_boundary
         Diagram.from_pyzx(graph)
     graph.auto_detect_io()
+    graph.set_inputs(graph.inputs() + graph.outputs())
     with raises(ValueError):  # duplicate_boundary
         Diagram.from_pyzx(graph)
 
 
-def test_backnforth_pyzx():
+def test_backnforth_pyzx_1():
     from pyzx import Graph
     path = 'test/src/zx-graph.json'
     graph = Graph.from_json(open(path).read())
@@ -122,21 +136,32 @@ def test_backnforth_pyzx():
     assert backnforth(diagram) == diagram
 
 
+def test_backnforth_pyzx_2(random_had_cnot_diagram):
+    from pyzx import compare_tensors
+    backnforth = lambda diagram: Diagram.from_pyzx(diagram.to_pyzx())
+    c = random_had_cnot_diagram(3, 25)
+    diagram = circuit2zx(c)
+    new_diagram = backnforth(diagram)
+    assert compare_tensors(
+        new_diagram.to_pyzx(), diagram.to_pyzx(), preserve_scalar=False
+    )
+
+
 def _std_basis_v(*c):
     v = np.zeros(2**len(c), dtype=complex)
     v[np.sum((np.array(c) != 0) * 2**np.arange(len(c)))] = 1
     return np.expand_dims(v, -1)
 
 
-def test_circui2zx():
+def test_circuit2zx():
     circuit = Ket(0, 0) >> quantum.H @ Rx(0) >> CRz(0) >> CRx(0) >> CU1(0)
-    circuit2zx(circuit) == Diagram(
+    assert circuit2zx(circuit) == Diagram(
         dom=PRO(0), cod=PRO(2), boxes=[
-            X(0, 1), X(0, 1), H, X(1, 1),
+            X(0, 1), X(0, 1), scalar(0.5), H, X(1, 1),
             Z(1, 2), Z(1, 2), X(2, 1), Z(1, 0),
             X(1, 2), X(1, 2), Z(2, 1), X(1, 0),
             Z(1, 2), Z(1, 2), X(2, 1), Z(1, 0)],
-        offsets=[0, 1, 0, 1, 0, 2, 1, 1, 0, 2, 1, 1, 0, 2, 1, 1])
+        offsets=[0, 1, 2, 0, 1, 0, 2, 1, 1, 0, 2, 1, 1, 0, 2, 1, 1])
 
     # Verify XYZ=iI
     t = circuit2zx(quantum.Z >> quantum.Y >> quantum.X)
@@ -155,3 +180,8 @@ def test_circui2zx():
     assert np.isclose(np.linalg.norm(t), 0)
     t = circuit2zx(Ket(0, 0)).to_pyzx().to_matrix() - _std_basis_v(0, 0)
     assert np.isclose(np.linalg.norm(t), 0)
+
+    assert (circuit2zx(quantum.Id(3).CX(0, 2))
+            == Diagram(dom=PRO(3), cod=PRO(3),
+                       boxes=[SWAP, Z(1, 2), X(2, 1), scalar(2 ** 0.5), SWAP],
+                       offsets=[1, 0, 1, 2, 1]))
